# Library of dictionary functions to use with SBL and similar algorithms
#
# file:         dictionary_functions.py
# Author:       Jakob Moederl
# Organization: Institute of Communication Networks and Satellite Communications
#               Graz University of Technology, Graz, Austria
# Date:         19.02.2025
# Version:      0.1

import numpy as np
from matplotlib.scale import scale_factory
from numpy import exp, sqrt, sin, arctan2, reshape
from numpy.linalg import norm
from scipy.linalg import khatri_rao
from scipy.constants import pi, c

from collections import namedtuple

LA = namedtuple('LA', ['origin', 'orientation', 'pos'])
def arrayresponse_radar_la_2d(pos, LA, f):
  """
  Array response of a MIMO Radar with virtual array given by LA and sample points (in frequency domain) given by f.

  :param pos: (K x 2) array of XY positions for which the array response should be evaluated
  :param LA: struct describing the LA. Must contain origin, orientation, and a vector of antenna positions relative to the origin
  :param f: vector of sample points of the (baseband) received signal
  :return: (N x K) matrix where each column corresponds to the (stacked) array response to a target located at pos
  """
  if pos.ndim == 1 and pos.size == 2:
    pos = pos.reshape([1, -1])
  A_angle = array_response_la(pos, LA)
  A_delay = array_response_delay(pos, LA.origin, f)
  return khatri_rao(A_angle, A_delay)

def arrayresponse_radar_la_2d_pathloss(pos, LA, f, d_unit=10, d_min=1):
  """
  Array response of a MIMO Radar with virtual array given by LA and sample points (in frequency domain) given by f.

  :param pos: (K x 2) array of XY positions for which the array response should be evaluated
  :param LA: struct describing the LA. Must contain origin, orientation, and a vector of antenna positions relative to the origin
  :param f: vector of sample points of the (baseband) received signal
  :return: (N x K) matrix where each column corresponds to the (stacked) array response to a target located at pos
  """
  if pos.ndim == 1 and pos.size == 2:
    pos = pos.reshape([1, -1])
  A_angle = array_response_la(pos, LA)
  A_delay = array_response_delay_pathloss(pos, LA.origin, f, d_unit, d_min)
  return khatri_rao(A_angle, A_delay)

def array_response_la(pos, LA):
  """
  Array response of a linear array given LA
  :param pos: XY-positions at which to evaluate the array response
  :param LA: struct describing the LA. Must contain .origin, .orientation and a vector .pos of antenna displacements
  :return: (N x K) dictionary matrix with N being the number antenna elements and K the number of evaluated positions
  """
  if pos.ndim == 1:
    pos = pos.reshape([1, -1])
  rel_pos = pos-np.ones([pos.shape[0], 1]) @ LA.origin.reshape(1, -1)
  angle = arctan2(rel_pos[:, 1], rel_pos[:, 0]) - LA.orientation
  return  1/sqrt(LA.pos.size) * exp(-2j*pi*LA.pos.reshape(-1, 1)*sin(angle.reshape(1, -1)))


def array_response_delay(pos, antenna_pos, f):
  """
  Array response of a monostatic radar with the received signal sampled (in frequency domain) at f
  :param pos: XY-positions at which to evaluate the array response
  :param antenna_pos: postion of the radar
  :param f: vector of sample points of the (baseband) received signal
  :return: (N x K) dictionary matrix with N being the number frequency points and K the number of evaluated positions
  """
  if pos.ndim == 1:
    pos = pos.reshape([1, -1])
  rel_pos = pos-np.ones([pos.shape[0], 1]) @ np.asarray(antenna_pos).reshape(1, -1)
  delay = 2*norm(rel_pos, axis=1)/c # we consider radar, therefor we have to count the two-way propagation delay
  return  1/sqrt(f.size) * exp(-2j*pi*f.reshape(-1, 1)*delay.reshape(1, -1))

def array_response_delay_pathloss(pos, antenna_pos, f, d_unit=10, d_min=1):
  """
  Array response of a monostatic radar with the received signal sampled (in frequency domain) at f, considering path loss
  :param pos: XY-positions at which to evaluate the array response
  :param antenna_pos: postion of the radar
  :param f: vector of sample points of the (baseband) received signal
  :param d_unit: distance at which the path loss is normalized to unity.
  :param d_min: cap for pathloss/path gain to avoid infinity at dist=0.
  :return: (N x K) dictionary matrix with N being the number frequency points and K the number of evaluated positions
  """
  if pos.ndim == 1:
    pos = pos.reshape([1, -1])
  rel_pos = pos-np.ones([pos.shape[0], 1]) @ np.asarray(antenna_pos).reshape(1, -1)
  dist = norm(rel_pos, axis=1)
  delay = 2*dist/c # we consider radar, therefor we have to count the two-way propagation delay
  dist[dist<d_min] = d_min
  scale_factor = (d_unit/dist)**2
  return  scale_factor/sqrt(f.size) * exp(-2j*pi*f.reshape(-1, 1)*delay.reshape(1, -1))

def lse_dictionary(theta, t=np.linspace(0, 1, 100, endpoint=True)):
  """
  dictionary generator function for (simple) LSE estimation
  :param theta: (normalized) frequency of the dictionary elements
  :param t: vector of sample times
  :return: matrix with spectral lines at frequencies theta_k for each column
  """
  return 1/sqrt(t.size) * exp(1j*2*pi * reshape(t, [-1,  1]) @ reshape(theta, shape=(1, -1)))