# Example plots for MIMO radar (Project with Anders Malte Westerkam from AAU)
##
# file:         sbl4radar_example.py
# Author:       Jakob Moederl
# Organization: Institute of Communication Networks and Satellite Communications
#               Graz University of Technology, Graz, Austria
# Date:         19.02.2025
# Version:      0.1

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm

from numpy import zeros, ones, exp, linspace, meshgrid, ceil, array, reshape, sqrt, abs
from scipy.constants import pi, c
from numpy.random import rand, randn

from pySBL import SBL
from pySBL.util import pow2db, db2pow, mag2db, db2mag
from pyOMP import OMP
from dictionary_functions import arrayresponse_radar_la_2d, LA
from gospa import gospa_dist

# Radar Array
radar1 = LA(pos=array([-1.5, -0.5, 0.5, -1, 0, -1, -0.5, 0.5, 1.5]), orientation=pi/2, origin=array([0, 0]))
BW = 20e6
d_max = 200
# this selects an odd number for N
N = int(ceil((d_max/c)*BW /2)*2+1)
# calculate sample vector
f = linspace(-BW/2, BW/2, N, endpoint=True)


SNR_dB = 30
signal_scaling = db2mag(SNR_dB)

OMP_threshold_dB = 7
OMP_threshold = db2pow(OMP_threshold_dB)
SBL_threshold_dB = 7.4

t_full = np.arange(-30, 31)

theta_t0 = np.array([[0, 30], [0.4, 30.3]])
delta_theta = np.array([[1.5, 0.5], [1.5, -0.5]])

theta_full = np.zeros((t_full.size, 2, 2))
for idx_t, t in enumerate(t_full):
  theta_full[idx_t, :, :] = theta_t0 + t * delta_theta

theta_est_sbl = []
theta_est_omp = []

H = lambda a: np.conj(a).transpose()
Pb = lambda y: abs(H(A_grid) @ y)**2


X_grid = linspace(-50, 50, 101, endpoint=True)
Y_grid = linspace(0, 100, 101, endpoint=True)

XX, YY = meshgrid(X_grid, Y_grid)
theta_grid = np.hstack([XX.reshape([-1, 1]), YY.reshape([-1, 1])])

A_fun = lambda theta: arrayresponse_radar_la_2d(theta, radar1, f)
A_grid = A_fun(theta_grid)


for idx_t, t in enumerate(t_full):
    print('------------------------------')
    print(f't = {t}')
    theta_true = theta_full[idx_t, :, :]
    x_true = signal_scaling * ones([theta_true.shape[0]])*exp(2j*pi*rand(theta_true.shape[0]))

    y_clean = A_fun(theta=theta_true) @ x_true
    y_noisy = y_clean + (randn(*y_clean.shape, 2) @ np.array([1, 1j])/sqrt(2))

    omp_res = OMP(y_noisy, A_fun=A_fun, theta_grid=theta_grid, A_grid=A_grid, threshold=OMP_threshold)
    theta_est_omp.append(omp_res.theta)
    print('Estimate OMP:')
    print(f'{omp_res.theta}')

    sbl = SBL(A_fun, theta_grid, threshold=db2pow(SBL_threshold_dB))
    sbl.run(y_noisy, init_state='empty')
    theta_est_sbl.append(sbl.theta)
    print('Estimate SBL:')
    print(f'{sbl.theta}')


theta_plot_omp = np.vstack(theta_est_omp)
theta_plot_sbl = np.vstack(theta_est_sbl)

fig, ax = plt.subplots()
ax.plot(theta_plot_sbl[:, 0], theta_plot_sbl[:, 1], 'ro', markerfacecolor='none', label='SBL')
ax.plot(theta_plot_omp[:, 0], theta_plot_omp[:, 1], 'b+', markerfacecolor='none', label='NOMP')
ax.plot(theta_full[:, :, 0].flatten(), theta_full[:, :, 1].flatten(), 'ko', markerfacecolor='k', markersize=3, label='True')
plt.grid()
plt.xlabel('x in m')
plt.ylabel('y in m')
plt.legend()
ax.axis('equal')
plt.show()


