# Example plots for MIMO radar (Project with Anders Malte Westerkam from AAU)
##
# file:         sbl4radar_example.py
# Author:       Jakob Moederl
# Organization: Institute of Communication Networks and Satellite Communications
#               Graz University of Technology, Graz, Austria
# Date:         19.02.2025
# Version:      0.1

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm

from numpy import zeros, ones, exp, linspace, meshgrid, ceil, array, reshape, sqrt, abs
from scipy.constants import pi, c
from numpy.random import rand, randn

from pySBL import SBL
from pySBL.util import pow2db, db2pow, mag2db, db2mag
from pyOMP import OMP
from dictionary_functions import arrayresponse_radar_la_2d, LA, arrayresponse_radar_la_2d_pathloss
from gospa import gospa_dist

# Radar Array
all_radars = [LA(pos=np.array([-1.5, -0.5, 0.5, -1, 0, -1, -0.5, 0.5, 1.5]), orientation=float(-pi / 2), origin=np.array([0,60])),
              LA(pos=np.array([-1.5, -0.5, 0.5, -1, 0, -1, -0.5, 0.5, 1.5]), orientation=float(pi), origin=np.array([30, 30])),
              LA(pos=np.array([-1.5, -0.5, 0.5, -1, 0, -1, -0.5, 0.5, 1.5]), orientation=float(0), origin=np.array([-30, 30])),
              LA(pos=np.array([-1.5, -0.5, 0.5, -1, 0, -1, -0.5, 0.5, 1.5]), orientation=float(pi / 2), origin=np.array([0, 0]))]
all_radars = all_radars[0:4]
La = len(all_radars) # Number of different dictionaries/radars
Ls = 1 # number of snapshots per radar

BW = 20e6
d_max = 200
# this selects an odd number for N
N = int(ceil((d_max/c)*BW /2)*2+1)
# calculate sample vector
f = linspace(-BW/2, BW/2, N, endpoint=True)

SNR_dB = 60
signal_scaling = db2mag(SNR_dB)

X_grid = linspace(-30, 30, 61, endpoint=True)
Y_grid = linspace(0, 60, 61, endpoint=True)

XX, YY = meshgrid(X_grid, Y_grid)
theta_grid = np.hstack([XX.reshape([-1, 1]), YY.reshape([-1, 1])])

A_fun_single = lambda theta, radar: arrayresponse_radar_la_2d_pathloss(theta, radar, f, d_unit=10, d_min=1)
# A_fun_single = lambda theta, radar: arrayresponse_radar_la_2d(theta, radar, f)
A_fun_multi = lambda theta: np.stack([A_fun_single(theta, radar) for radar in all_radars], axis=0)
A_grid = A_fun_multi(theta_grid)

theta_true = np.array([[0, 10], [-20, 30], [0, 50], [20, 30]])
x_true = signal_scaling * ones([La, theta_true.shape[0], Ls])*exp(2j*pi*rand(La, theta_true.shape[0], Ls))

y_clean = A_fun_multi(theta=theta_true) @ x_true
y_noisy = y_clean + (randn(*y_clean.shape, 2) @ np.array([1, 1j])/sqrt(2))

H = lambda a: np.conj(a).swapaxes(-2, -1)
Pb = lambda y: np.mean(abs(np.squeeze(H(A_grid) @ y))**2/np.linalg.norm(A_grid, axis=1)**4, axis=0)

Pb_data = Pb(y_noisy)
Pb_data_dB = pow2db(Pb_data)

OMP_threshold = 5
OMP_threshold_dB = pow2db(OMP_threshold)
SBL_threshold_dB = 7.2

# omp_res = OMP(y_noisy, A_fun=A_fun, theta_grid=theta_grid, A_grid=A_grid, threshold=OMP_threshold)
# print('Estimate OMP:')
# print(f'{omp_res.theta}')

sbl = SBL(A_fun_multi, theta_grid, threshold=db2pow(SBL_threshold_dB))
sbl.run(y_noisy)
print('Estimate SBL:')
print(f'{sbl.theta}')

ospa_sbl = gospa_dist(sbl.theta.transpose(), theta_true.transpose(), c=10, p=2)
# ospa_omp = gospa_dist(omp_res.theta.transpose(), theta_true.transpose(), c=10, p=2)

print(f'OSPA SBL: {ospa_sbl}')

fig, ax = plt.subplots(subplot_kw={'projection': '3d'})
surf = ax.plot_surface(theta_grid[:, 0].reshape([Y_grid.size, X_grid.size]),
                       theta_grid[:, 1].reshape([Y_grid.size, X_grid.size]),
                       Pb_data.reshape([Y_grid.size, X_grid.size]),
                       cmap=cm.viridis)
plt.xlabel('x in m')
plt.ylabel('y in m')
ax.set_zlabel('Bartlet Power (linear)')

fig, ax = plt.subplots()
im = ax.imshow(Pb_data_dB.reshape(Y_grid.size, X_grid.size), origin='lower', extent=[min(X_grid), max(X_grid), min(Y_grid), max(Y_grid)], cmap=cm.viridis, vmax=np.max(Pb_data_dB), vmin=np.max(Pb_data_dB)-40)
ax.plot(theta_true[:, 0], theta_true[:, 1], 'rs', markerfacecolor='none', label='true')
ax.plot(sbl.theta[:, 0], sbl.theta[:, 1], 'r+', markerfacecolor='none', label='SBL')
# ax.plot(omp_res.theta[:, 0], omp_res.theta[:, 1], 'ro', markerfacecolor='none')
cbar = fig.colorbar(im)
cbar.set_label('Barlett Power in dB')
plt.xlabel('x in m')
plt.ylabel('y in m')
plt.legend()
plt.show()


