# Example plots for MIMO radar (Project with Anders Malte Westerkam from AAU)
##
# file:         sbl4radar_example.py
# Author:       Jakob Moederl
# Organization: Institute of Communication Networks and Satellite Communications
#               Graz University of Technology, Graz, Austria
# Date:         19.02.2025
# Version:      0.1

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm

from numpy import zeros, ones, exp, linspace, meshgrid, ceil, array, reshape, sqrt, abs
from scipy.constants import pi, c
from numpy.random import rand, randn

from pySBL import SBL
from pySBL.util import pow2db, db2pow, mag2db, db2mag
from pyOMP import OMP
from dictionary_functions import arrayresponse_radar_la_2d, LA
from gospa import gospa_dist

# Radar Array
radar1 = LA(pos=array([-1.5, -0.5, 0.5, -1, 0, -1, -0.5, 0.5, 1.5]), orientation=pi/2, origin=array([0, 0]))
BW = 20e6
d_max = 200
# this selects an odd number for N
N = int(ceil((d_max/c)*BW /2)*2+1)
# calculate sample vector
f = linspace(-BW/2, BW/2, N, endpoint=True)

SNR_dB = 15
signal_scaling = db2mag(SNR_dB)

X_grid = linspace(-100, 100, 201, endpoint=True)
Y_grid = linspace(0, 100, 101, endpoint=True)

XX, YY = meshgrid(X_grid, Y_grid)
theta_grid = np.hstack([XX.reshape([-1, 1]), YY.reshape([-1, 1])])

A_fun = lambda theta: arrayresponse_radar_la_2d(theta, radar1, f)
A_grid = A_fun(theta_grid)

theta_true = np.array([[0, 20], [-20, 50]])
x_true = signal_scaling * ones([theta_true.shape[0]])*exp(2j*pi*rand(theta_true.shape[0]))

y_clean = A_fun(theta=theta_true) @ x_true
y_noisy = y_clean + (randn(*y_clean.shape, 2) @ np.array([1, 1j])/sqrt(2))

H = lambda a: np.conj(a).transpose()
Pb = lambda y: abs(H(A_grid) @ y)**2

Pb_data = Pb(y_noisy)

OMP_threshold = 5
OMP_threshold_dB = pow2db(OMP_threshold)
SBL_threshold_dB = 7.2

omp_res = OMP(y_noisy, A_fun=A_fun, theta_grid=theta_grid, A_grid=A_grid, threshold=OMP_threshold)
print('Estimate OMP:')
print(f'{omp_res.theta}')

sbl = SBL(A_fun, theta_grid, threshold=db2pow(SBL_threshold_dB))
sbl.run(y_noisy)
print('Estimate SBL:')
print(f'{sbl.theta}')

ospa_sbl = gospa_dist(sbl.theta.transpose(), theta_true.transpose(), c=10, p=2)
ospa_omp = gospa_dist(omp_res.theta.transpose(), theta_true.transpose(), c=10, p=2)

print(f'OSPA SBL: {ospa_sbl}, OSPA OMP: {ospa_omp}')

fig, ax = plt.subplots(subplot_kw={'projection': '3d'})
surf = ax.plot_surface(theta_grid[:, 0].reshape([Y_grid.size, X_grid.size]),
                       theta_grid[:, 1].reshape([Y_grid.size, X_grid.size]),
                       Pb_data.reshape([Y_grid.size, X_grid.size]),
                       cmap=cm.viridis)
plt.xlabel('x in m')
plt.ylabel('y in m')
ax.set_zlabel('Bartlet Power (linear)')

fig, ax = plt.subplots()
im = ax.imshow(Pb_data.reshape(Y_grid.size, X_grid.size), origin='lower', extent=[min(X_grid), max(X_grid), min(Y_grid), max(Y_grid)], cmap=cm.viridis)
ax.plot(theta_true[:, 0], theta_true[:, 1], 'rs', markerfacecolor='none', label='true')
ax.plot(sbl.theta[:, 0], sbl.theta[:, 1], 'r+', markerfacecolor='none', label='SBL')
ax.plot(omp_res.theta[:, 0], omp_res.theta[:, 1], 'ro', markerfacecolor='none', label='OMP')
cbar = fig.colorbar(im)
cbar.set_label('Barlett Power in dB')
plt.xlabel('x in m')
plt.ylabel('y in m')
plt.legend()
plt.show()


