# Computation of generalized optimal subpatter assignment (GOSPA) metric presented in [1]
#
#   [1] A. S. Rahmathullah, Á. F. García-Fernández, and L. Svensson, "Generalized optimal sub-pattern assignment metric"
#   in 20th International Conference on Information Fusion, Xi'an, China, July 10-13, 2017,
#   doi: 10.23919/ICIF.2017.8009645
#
# file:         gospa.py
# Author:       Jakob Moederl
# Organization: Institute of Communication Networks and Satellite Communications
#               Graz University of Technology, Graz, Austria
# Date:         19.02.2025
# Version:      0.1

import numpy as np
from numpy import sum, tile, ones, kron, abs
from numpy.linalg import norm
from munkres import Munkres

euclid_dist = lambda x: norm(x, axis=0)

def gospa_dist(X, Y, c, p=2, costfun=euclid_dist):
  """
  calculate Optimal Sub-Pattern Assignment (OSPA) Metric
  Assume that each column-vector in X and Y are
  """

  X = np.array(X)
  Y = np.array(Y)
  if X.size == 0 and Y.size == 0:
    return 0
  elif X.size == 0 or Y.size == 0:
    return c

  M = X.shape[1]
  N = Y.shape[1]
  # calculate cost matrix for pairings
  XX = tile(X, [1, N])
  YY = kron(Y, ones([1, M]))
  D = costfun(XX-YY).reshape([N, M])
  D[D>c] = c
  if D.shape[1] < D.shape[0]:
    D = np.hstack([D, c*ones([D.shape[0], D.shape[0]-D.shape[1]])])
  elif D.shape[1] > D.shape[0]:
    D = np.vstack([D, c*ones([D.shape[1]-D.shape[0], D.shape[1]])])

  # find best assignment
  m = Munkres()
  assignments = m.compute(D.copy())

  return (sum([D[ii, jj]**p for ii, jj in assignments])/max(M, N))**(1/p)