# function OMP_FMIN orthogonal matching pursuit algorithm for parameterized dictionaries with numeric optimization
#   Note: This algorithm assume (complex) AWGN and that the signal is normalized such that the noise has unit variance.
#
#   res = OMP(y, A_fun, theta_grid, A_grid, threshold, maxIter) applies the OMP-NO algorithm to (complex-valued) data y
#   generated by
#       y = \sum_{k=1}^{K} A(theta_k) x_k + v
#   where A(theta_k) are signa atoms paramterized by paramters theta_k generated using the function A_fun(theta_k),
#   which takes in the (1 x L)-tuple of signal atom parameters theta_k, in additive white gaussian noise with unit
#   variance (sigma^2 = 1)
#
#   The result obtained is a namedtuple with entries theta (K x L) and x (K x 1).
#
#   The algorithm in this file is based on [1], except it uses scipy.optimze instead of newton updates.
#
#   [1] B. Mamandipoor, D. Ramasamy, and U. Madhow, “Newtonized orthogonal matching pursuit: Frequency estimation over
#   the continuum,” IEEE Trans. Signal Process., vol. 64, no. 19, pp. 5066–5081, Oct. 2016,
#   doi: 10.1109/TSP.2016.2580523
#
# file:         pyOMP.py
# Author:       Jakob Moederl
# Organization: Institute of Communication Networks and Satellite Communications
#               Graz University of Technology, Graz, Austria
# Date:         19.02.2025
# Version:      0.1



import numpy as np
from numpy import abs, conj, swapaxes, mean
from numpy.linalg import norm
from scipy import optimize as sciop
from collections import namedtuple

OMPState = namedtuple('OMPstate', ['theta', 'x'])

H = lambda x: conj(swapaxes(x, -2, -1)) if x.ndim > 1 else conj(x)

def OMP(y, A_fun, theta_grid, A_grid=None, threshold=1, maxIter=50, Nrefinement=1):
  """
  Orthogonal matching pursuit (OMP) for parameterized dictionaries using numeric optimization (scipy.optimize) for
  continuous parameter estimates.

  Attention: This algorithm assumes AWGN and that the data y is scaled such that the AWGN has unit variance!

  :param y: (N x 1) noisy data
  :param A_fun: function handle that takes as input a (K x L) parameter array and returns a (N x K) dictionary
  :param theta_grid: (P x L) search grid of parameters to aid the search for new components
  :param A_grid: (N x P) dictionary evaluated on the initial search grid
  :param threshold: threshold value after at which the algorithm stopps adding new components and terminates
  :param maxIter: maximum number of iterations
  :return: namedtuple with parameter estimates theta (K x L) and amplitudes x (K x 1)
  """
  if y.ndim == 1:
    y = y.reshape(-1, 1)

  if A_grid is None:
    A_grid = A_fun(theta_grid)

  if theta_grid.ndim == 1:
    theta_grid = theta_grid.reshape(-1, 1)
  y_res = y
  x = np.array([]).reshape(-1, y.shape[1])
  theta = np.array([]).reshape(0, theta_grid.shape[1])

  logP_single = lambda y_res, A_test: mean(abs(H(A_test) @ y_res) ** 2) / (norm(A_test)**2)


  for iter in range(maxIter):
    Q_grid = mean(abs(H(A_grid) @ y_res)**2, axis=1) / (norm(A_grid, axis=0)**2)
    idx_Q_max = Q_grid.argmax()
    theta_hat_new = theta_grid[idx_Q_max, :]
    Q_max = Q_grid[idx_Q_max]

    optim_res = sciop.minimize(lambda x: -logP_single(y_res, A_fun(x)), theta_hat_new)
    Q_opt = -optim_res.fun

    # sanity check and either keep grid-based value or value from optimization
    if Q_opt > Q_max:
      theta_hat_new = optim_res.x
      Q_max = Q_opt

    # check exit condition
    if Q_max < threshold:
      break

    theta = np.vstack([theta, theta_hat_new])
    Ak = A_fun(theta_hat_new)
    xk = (norm(Ak)**(-2)) * H(Ak) @ y_res
    x = np.vstack([x, xk])
    y_res = y_res - Ak @ xk

    for _ in range(Nrefinement):
      for jj in range(theta.shape[0]):
        Ak_old = A_fun(theta[jj, :])
        y_res = y_res + Ak_old @ x[jj, :].reshape(1, -1)
        Q_old = logP_single(y_res, Ak_old)

        optim_res = sciop.minimize(lambda x: -logP_single(y_res, A_fun(x)), theta[jj, :])
        Q_opt = -optim_res.fun

        # sanity check
        if Q_opt > Q_old:
          theta[jj, :] = optim_res.x
          Ak = A_fun(theta[jj, :])
          x[jj, :] = (norm(Ak)**(-2)) * H(Ak) @ y_res
        else: # no improvement, skip update
          Ak = Ak_old
        y_res = y_res - Ak @ x[jj, :].reshape(1, -1)
      # end loop over existing components
    # end loop over refinement iterations
  # end algorithm main loop

  return OMPState(theta=theta, x=x)






