# Super-resolution SBL with 3D-Data-Tensor
# Author: Jakob Moederl (jakob.moederl@tugraz.at)
#   Graz University of Technology, Austria
# Version: 0.1
# Date: December 2024
#
# SBL implementation with 3D-Data-Tensor (Samples x Snapshots x Measurement Appertures), i.e. considering we have L_F
# appertures observing the signal with different dictionaries. For each appertures we obtain L_S independent snapshots
# containing N samples each. The dictionary parameters are assumed to factorize (i.e. each subset of parameters acts
# only on a single column of each dictionary) and estimated parametrically


import numpy as np
from numpy import pi, exp, log, log10, sqrt, min, max, abs, angle, real, imag, sum, mean, any, reshape, squeeze, zeros, ones
from numpy.polynomial import Polynomial
from scipy import optimize as sciop
from scipy.linalg import convolution_matrix

from collections import namedtuple
import warnings

from .util import Es, H

SBLState = namedtuple('SBLState', ['theta', 'gamma', 'lam', 'x', 'SNR'])

class SBL:
  def __init__(self, dict_fun, theta_grid,
               lambda_init=None, max_iter=50, convergence_threshold=1e-4, gamma_prior=0, threshold=1, keep_history=True):
    """
    Initialize new super-resolution SBL estimator object
    :param dict_fun: function handel to generate the dictionary columns
    :param theta_grid: M x Ntheta matrix of initial search grid of theta values (for initializing new components)
    :param lambda_init: initial value of the noise precision estimate
    :param max_iter: maximum number of iterations
    :param gamma_prior: value of a in the prior varphi(gamma)=exp(-rho (N Ls La) a / gamma)
    :param threshold: threshold of component SNR (linear)
    """

    if theta_grid.ndim==1:
      theta_grid = reshape(theta_grid, [-1, 1])
    self._theta_grid = theta_grid

    A_test = dict_fun(theta_grid[0, :])
    if A_test.ndim == 2:
      self.La = 1
      self.N = A_test.shape[0]
      self.Ls = 1
      self._dict_fun = lambda theta: reshape(dict_fun(theta), (self.La, self.N, -1))
    elif A_test.ndim == 3:
      self.La, self.N, _ = A_test.shape
      self.Ls = 1
      self._dict_fun = dict_fun
    else:
      raise RuntimeError('dict_fun must be a function which returns a 2-dimensional (N x K) array or a 3-dimensional (La x N x K) array, where N is the number of samples per observations and La is the number of apertures (i.e. different dictionaries)')

    self._dict_grid = None
    self.max_iter = max_iter
    self.gamma_prior = gamma_prior
    self.threshold = threshold
    self.lambda_init = lambda_init
    self.keep_history = keep_history
    self.convergence_threshold = convergence_threshold
    self._results = []

  def run(self, y, init_state=None):
    """
    Run SBL estimator on given data
    :param y: data tensor of size (La x N x Ls), (La x N), (N x Ls), (N)
    :param init_state: initial state of the SBL estimator (either None, 'last', lambda_init or tuple of (theta, gamma, lambda))')
    :return: namedtuple('theta', 'gamma', 'lam', 'x', 'SNR') of final iteration if successful, None otherwise.
    """

    if y.ndim == 1:
      self.Ls = 1
      y = reshape(y, (self.La, self.N, self.Ls))
    elif y.ndim == 2:
      if y.shape[0] == self.La and y.shape[1] == self.N:
        y = reshape(y, (self.La, self.N, self.Ls))
      else:
        self.Ls = y.shape[-1]
        y = reshape(y, (self.La, self.N, self.Ls))

    #########################################
    # Initialization                        #
    #########################################
    x_init = zeros((self.La, 0, self.Ls))
    SNR_init = zeros(0)
    if (init_state is None) or (init_state == 'last' and len(self.history) == 0):
      # No init state given
      lambdaHat = 10 * self.Ntotal / (Es(y))
      theta = zeros((0, self.theta_grid.shape[1]))
      gamma = zeros(0)
    elif init_state == 'last':
      # use previous state as init
      theta, gamma, lambdaHat, x_init, SNR_init = self.res
    elif init_state == 'empty':
      lambdaHat = 10 * self.Ntotal / (Es(y))
      theta = zeros((0, self.theta_grid.shape[1]))
      gamma = zeros(0)
    else:
      # try to use interpret given state as noise variance
      try:
        lambdaHat = float(init_state)
        theta = zeros((0, self.theta_grid.shape[1]))
        gamma = zeros(0)
      except TypeError:
        pass
      # try to use given state as init
      try:
        # initialize theta, gamma and lambda (necessary)
        theta, gamma, lambdaHat, *_ = init_state
        try:
          # try to also initialize x_hat and component SNRs
          x_init = init_state[3]
          SNR_init = init_state[4]
        except IndexError:
          # Failed to initialized x_hat and component SNRs -- ignore
          pass
      except TypeError:
        # Failed to initialize theta, gamma and lambda -- raise error
        raise RuntimeError("SBL init state must be either None, 'last' or a tuple containing (at minimum) theta, gamma, lambdaHat")
    current_state = SBLState(theta, gamma, lambdaHat, x_init, SNR_init)

    ######################################
    # Main loop                          #
    ######################################
    for idx_iter in range(self.max_iter):
      old_state = current_state

      ##########################################################
      # loop over already existing components to optimize them #
      ##########################################################

      # list of component SNRs for logging
      component_snr = []
      # boolean mask to track which components get removed
      valid_other_mask = np.ones((theta.shape[0]), dtype=np.bool)
      for idx_comp in range(theta.shape[0]):
        valid_other_mask[idx_comp] = False
        theta_k = theta[idx_comp, :]
        gamma_k = gamma[idx_comp]
        theta_other = theta[valid_other_mask, :]
        gamma_other = gamma[valid_other_mask]

        # pre-calculate neccessary matrices
        if gamma_other.size == 0:
          B = zeros(shape=(self.La, self.N, 0))
          y_res = y
        else:
          A_kbar = self._dict_fun(theta_other)
          L = get_covar_factorization(A_kbar, gamma_other, lambdaHat)
          B = A_kbar @ L
          y_res = y - lambdaHat * B @ (H(B) @ y)


        # optimze objective function
        if self.La == 1:
          QoverS = lambda Q, s: Q/s
          opt_fun = lambda x: -QoverS(*get_Qs(y_res, self._dict_fun(x), lambdaHat, B))
        else:
          opt_fun = lambda x: -sbl_objective(*get_Qs(y_res, self._dict_fun(x), lambdaHat, B), gamma_k)
        opt_res = sciop.minimize(opt_fun, theta_k)
        theta_k_new = opt_res.x

        # update gamma
        Q_k, s_k = get_Qs(y_res, self._dict_fun(theta_k_new), lambdaHat, B)
        gamma_k_new = solve_gamma_poly(Q_k, s_k, prior=self.gamma_prior, threshold=self.threshold)
        if gamma_k_new:
          theta[idx_comp, :] = theta_k_new
          gamma[idx_comp] = gamma_k_new
          component_snr.append(mean(Q_k/s_k))
          # re-set mask so that the component is kep in the model
          valid_other_mask[idx_comp] = True
        # else:
        #   do not re-enable the component for the next iteration of the loop
        #   pass
      # END OF FOR LOOP OVER COMPONENTS
      # Apply mask to remove non-valid compoents from the arrays
      theta = theta[valid_other_mask, :]
      gamma = gamma[valid_other_mask]
      ######################################
      # search for new component to add
      #####################################

      # pre-calculate necessary matrices
      if gamma.size == 0:
        B = zeros(shape=(self.La, self.N, 0))
        y_res = y
      else:
        A_kbar = self._dict_fun(theta)
        L = get_covar_factorization(A_kbar, gamma, lambdaHat)
        B = A_kbar @ L
        y_res = y - lambdaHat * B @ (H(B) @ y)

      # brute-force grid to search for maximum
      # gamma_hat_init = 1e-20
      Q_grid, s_grid = get_Qs(y_res, self.A_grid, lambdaHat, B)
      # if self.La == 1:
      #   objective_grid = Q_grid / s_grid
      # else:
        # objective_grid = sbl_objective_init(Q_grid, s_grid)
      objective_grid = np.mean(Q_grid / s_grid, axis=0)
      idx_max = np.argmax(objective_grid)

      # refine grid-based estimate
      # if self.La == 1:
      QoverS = lambda Q, s: np.mean(Q / s, axis=0)
      opt_fun = lambda x: -QoverS(*get_Qs(y_res, self._dict_fun(x), lambdaHat, B))
      # else:
      #   opt_fun = lambda x: -sbl_objective_init(*get_Qs(y_res, self._dict_fun(x), lambdaHat, B))
      opt_res = sciop.minimize(opt_fun, self.theta_grid[idx_max, :])
      theta_new = opt_res.x

      # check if component should be added
      Q_init, s_init = get_Qs(y_res, self._dict_fun(theta_new), lambdaHat, B)
      gamma_new = solve_gamma_poly(Q_init, s_init, prior=self.gamma_prior, threshold=self.threshold)
      if gamma_new:
        gamma = np.append(gamma, gamma_new)
        theta = np.vstack((theta, theta_new))
        component_snr.append(mean(Q_init/s_init))

      #############################################
      # Update noise precision lambda
      #############################################
      A = self._dict_fun(theta)
      L = get_covar_factorization(A, gamma, lambdaHat)
      B = A @ L
      x_hat = lambdaHat * L @ (H(L) @ (H(A) @ y))
      lambdaHat = self.Ntotal / (Es(y-A@x_hat) + Es(B))

      # Finalize State
      current_state = SBLState(theta, gamma, lambdaHat, x_hat, np.array(component_snr))

      if self.keep_history:
        self._results.append(current_state)
      else:
        self._results = [current_state]

      if gamma.size == 0:
        warnings.warn(RuntimeWarning("No components found"))
        return

      if is_converged(old_state, current_state, self.convergence_threshold):
        return current_state
    # END FOR LOOP OVER ALGORITHM ITERATIONS
    warnings.warn(RuntimeWarning(f'SBL Reached max_iter={self.max_iter} iterations without converging'))
    return current_state


  @property
  def Ntotal(self):
    return self.N * self.Ls * self.La if self.La else self.N * self.Ls

  @property
  def theta_grid(self):
      return self._theta_grid

  @theta_grid.setter
  def theta_grid(self, theta_grid):
    self._theta_grid = theta_grid
    self._dict_grid = None

  @property
  def A_grid(self):
    if self._dict_grid is None:
      self._dict_grid = self._dict_fun(self._theta_grid)
    return self._dict_grid

  @property
  def history(self):
    if not self.keep_history:
      warnings.warn(RuntimeWarning('Accessing SBL iteration history with <keep_history=false> will only return the final values'))
    return self._results

  @property
  def res(self):
    return self._results[-1]

  @property
  def theta(self):
    return self._results[-1].theta

  @property
  def gamma(self):
    return self._results[-1].gamma

  @property
  def K(self):
    return len(self.gamma)

  @property
  def lam(self):
    return self._results[-1].lam

  @property
  def yhat(self):
    return self._dict_fun(self.theta) @ self.res.x


def sbl_objective(Q, s, gamma):
  """
  SBL Objective function for individual component with marginal squared mean Q, variance s and component precision gamma
  :param Q:
  :param s:
  :param gamma:
  :return:
  """
  return squeeze(mean((Q / s) / (1+gamma * s) + log(s) + log(gamma) - log(1+gamma*s), axis=0))

def sbl_objective_init(Q, s):
  """
  SBL Objective function for initialization of new components with marginal squared mean Q, variance s
  :param Q:
  :param s:
  :return:
  """
  return squeeze(mean((Q / s) + log(s), axis=0))

def sbl_objective_init(Q, s):
  """
  SBL Objective function for initialization of new components with marginal squared mean Q, variance s
  :param Q:
  :param s:
  :return:
  """
  return squeeze(mean((Q / s) + log(s), axis=0))


def get_Qs(y_res, Ak, lambdaHat, B):
  """
  Calcualte marginal squared mean Q and variance of weights corresponding to dictionary entries Ak
  :param y_res: residual signal (y-lam*A_bar^H Sigma_bar_x A_bar y
  :param Ak: N x K matrix of dictionary columns to evaluate
  :param gamma_k: estimate of component precision
  :param lambdaHat: estimate of noise precision
  :param B: B=A_bar @ L where L is some decomposition such that Sigma_bar = L @ L^H
  :return: Q, s
  """
  s = 1/(lambdaHat * sum(abs(Ak) ** 2, axis=1) - (lambdaHat**2) * sum(abs(H(Ak) @ B) ** 2, axis=2))
  Q = (lambdaHat**2) * (s**2) * mean(abs(H(Ak) @ y_res) ** 2, axis=2)
  return Q, s

def get_covar_factorization(A, gammaEst, lambdaEst):
  """
  Returns a decomposition L of Sigma_bar = (lam * A_bar^H @ A_bar + diag(gamma)^(-1) such that Sigma_bar = L @ L^H
  :param A: La x N x K matrix of dictionary columns
  :param gammaEst: vector of corresponding prior precisions
  :param lambdaEst: noise precision estimate
  :return: K x K matrix L
  """
  La = A.shape[0]
  CovarInv = (lambdaEst * H(A) @ A) + np.tile(np.diag(gammaEst), (La, 1, 1))
  eigvals, eigvecs = np.linalg.eigh(CovarInv)
  if any(eigvals <= 0):
    raise RuntimeError('Got zero or negative eigenvalue in weight covariance matrix')
  eigvals = 1/sqrt(eigvals)
  return eigvecs @ np.apply_along_axis(func1d=np.diag, axis=1, arr=eigvals)

def is_converged(old_state, new_state, convergence_threshold=1e-4):
  """
  Check whether the SBL algorithm is converged based on difference of states
  :param old_state: state of previous iteration
  :param new_state: state of current interation
  :return: boolean indicating convergence
  """
  # check for recent addition/deletion of components
  if len(old_state.gamma) != len(new_state.gamma):
    return False
  # check of changes in gamma
  if max(abs(new_state.gamma - old_state.gamma)/abs(new_state.gamma + old_state.gamma)) > convergence_threshold:
    return False
  # check of changes in theta
  if (not np.any(abs(old_state.theta + new_state.theta) == 0) and
          max(abs(new_state.theta - old_state.theta)/abs(old_state.theta + new_state.theta)) > convergence_threshold):
    return False
  # check for convergence in lambda
  if abs(new_state.lam - old_state.lam)/abs(old_state.lam + new_state.lam) > convergence_threshold:
    return False
  # nothing important has changed, so I guess we are converged...
  return True

def solve_gamma_poly(Q, s, prior=0, threshold=1):
  """
  Get gamma candidates by solving the polynomial corresponding to the fixed-point equation
  :param Q: vector of squared means |q_k|^2
  :param s: vector of variances s
  :param prior: parameter a of the prior varphi(gamma)=exp(-rho a / gamma)
  :param threshold: threshold of component SNR (sum |q_k|^2/s_k) which must be exceeded
  :return: vector of possible gamma values
  """
  Q, s = squeeze(Q), squeeze(s)

  if mean(Q/s) <= threshold:
    return 0

  if Q.size == 1: # Single-Dictionary case
    if prior == 0: # Constant prio varphi(gamma)=const.
      gamma = 1/(Q-s) if Q/s > 1 else zeros(0)
      return gamma
    else: # Wipf's prior varphi(gamma) = exp(-rho M L a / gamma)
      beta = Q - s - prior * s ** 2
      alpha = 1/2 + prior*s
      gamma = (alpha + sqrt(alpha**2 + prior*beta)) / beta if beta > 0 and Q/s else zeros(0)
      return gamma
  else: # Multi-Dictionary case
    u = np.vstack((zeros(Q.size), ones(Q.size), -(Q-s), zeros((2*(Q.size-1), Q.size))))
    v = np.vstack((ones(Q.size), 2*s, s**2, zeros((2*(Q.size-1), Q.size))))

    B = u
    A = np.zeros(2*Q.size+1)
    A[0] = 1

    for idx in range(Q.size):
      idx_other = np.ones(Q.size, dtype=bool)
      idx_other[idx] = False
      convmtx = convolution_matrix(v[:, idx], 2*Q.size+1)
      convmtx = convmtx[:2*Q.size+1, :]
      B[:, idx_other] = convmtx @ B[:, idx_other]
      A = convmtx @ A
    B = 1/Q.size * sum(B, axis=1)

    if prior == 0:
      updatePoly = Polynomial(B[1:])
    else:
      updatePoly = Polynomial(B + prior*A)
    gamma_cand = updatePoly.roots()
    gamma_cand_real = real(gamma_cand[abs(angle(gamma_cand))<1e-10])

    if gamma_cand_real.size == 0:
      return 0
    elif gamma_cand_real.size == 1:
      return gamma_cand_real
    else:
      objective = [sbl_objective(Q, s, gamma_k) for gamma_k in gamma_cand_real]
      return gamma_cand_real[np.argmax(objective)]